open Util

open Ast

let parse filename =
  read_file filename begin fun chan ->
    let lexbuf = Lexing.from_channel chan in
    lexbuf.Lexing.lex_curr_p <- { lexbuf.Lexing.lex_curr_p with Lexing.pos_fname = filename };
    Parser.main Lexer.main lexbuf
  end

let variables_sz = 17
let generated_variables_sz = 17
let search_variables_sz = 17
let parameters_sz = 17

type var =
  | Unbound
  | Array of Expr.t list
  | Lin_eq of (float * Expr.t) list * float
  | Unary_op of string * Expr.t
  | Bin_op of Expr.t * string * Expr.t
  | Min of Expr.t * Expr.t
  | Max of Expr.t * Expr.t
  | Sqrt of Expr.t
  | Sin of Expr.t
  | Cos of Expr.t

let variables = Hashtbl.create variables_sz

type cstr =
  | Lin_lt of Expr.t * Expr.t * Expr.t
  | Lt of Expr.t * Expr.t
  | BoudingCircleLt of Expr.t * Expr.t * Expr.t * Expr.t
  | NonOverlapTriangles of Expr.t * Expr.t * Expr.t * Expr.t * Expr.t * Expr.t * Expr.t * Expr.t * Expr.t * Expr.t * Expr.t * Expr.t

let constraints = ref []

type output =
  | OutputVar of string
  | OutputArray of string * int * int

let output = ref []

let add_lin_lt x y z =
  constraints := Lin_lt (x, y, z) :: !constraints

let parameters = Hashtbl.create parameters_sz

let add_lin_eq x y z a p =
  let x =
    match x with
    | Expr.Var v ->
        Hashtbl.find parameters v
    | _ ->
        x in
  match x, y, z with
  | Expr.ArrayList coefs, Expr.ArrayList vars, Expr.Float constant ->
      let introduced =
        match a with
        | ["defines_var", [Expr.Var var]] -> var
        | _ -> Err.stop p "defines_var" in
      let coefs = List.map begin fun coef ->
        match coef with
        | Expr.Float f -> f
        | _ -> Err.stop p "add_lin_eq"
      end coefs in
      let introduced, inv_coef, vars, coefs =
        let rec find vars coefs =
          match vars, coefs with
          | [], [] ->
              raise Exit
          | Expr.Var var :: vars, inv_coef :: coefs when var = introduced ->
              inv_coef, vars, coefs
          | var :: vars, coef :: coefs ->
              let inv_coef, vars, coefs = find vars coefs in
              inv_coef, var :: vars, coef :: coefs
          | _ -> Err.stop p "coefs" in
        try
          let inv_coef, vars, coefs = find vars coefs in
          introduced, inv_coef, vars, coefs
        with Exit ->
          match vars, coefs with
          | Expr.Var var :: vars, inv_coef :: coefs ->
              var, inv_coef, vars, coefs
          | _ ->
              Err.stop p "defined var not found" in
      let vars = List.map2 begin fun coef x ->
        ((-. coef) /. inv_coef, x)
      end coefs vars in
      Hashtbl.replace variables introduced (Lin_eq (vars, constant /. inv_coef))
  | _ ->
      Err.stop p "add_lin_eq"

let add_lt x y =
  constraints := Lt (x, y) :: !constraints

let set_variable x v =
  match x with
  | Expr.Var x ->
      Hashtbl.replace variables x v
  | _ ->
      failwith "set_variable"

let add_times x y z =
  set_variable z (Bin_op (x, "*", y))

let add_min x y z =
  set_variable z (Min (x, y))

let add_max x y z =
  set_variable z (Max (x, y))

let add_sqrt x y =
  set_variable y (Sqrt x)

let add_sin x y =
  set_variable y (Sin x)

let add_cos x y =
  set_variable y (Cos x)

let add_bounding_circle_lt x y r l =
  constraints := BoudingCircleLt (x, y, r, l) :: !constraints

let add_non_overlap_triangles x0 y0 x1 y1 x2 y2 x3 y3 x4 y4 x5 y5 =
  constraints := NonOverlapTriangles (x0, y0, x1, y1, x2, y2, x3, y3, x4, y4, x5, y5) :: !constraints

let add_constraint c a p =
  match c with
  | ("float_lin_le", [x; y; z]) ->
      add_lin_lt x y z
  | ("float_lin_lt", [x; y; z]) ->
      add_lin_lt x y z
  | ("float_lin_eq", [x; y; z]) ->
      add_lin_eq x y z a p
  | ("float_lt", [x; y]) ->
      add_lt x y
  | ("float_times", [x; y; z]) ->
      add_times x y z
  | ("float_min", [x; y; z]) ->
      add_min x y z
  | ("float_max", [x; y; z]) ->
      add_max x y z
  | ("float_sqrt", [x; y]) ->
      add_sqrt x y
  | ("float_sin", [x; y]) ->
      add_sin x y
  | ("float_cos", [x; y]) ->
      add_cos x y
  | ("bool_clause", [_; _]) ->
      ()
  | ("array_bool_or", [_; _]) ->
      ()
  | ("float_eq_reif", [_; _; _]) ->
      ()
  | ("float_lin_eq_reif", [x; y; z; _]) ->
      add_lin_eq x y z a p
  | ("float_lin_lt_reif", [_; _; _; _]) ->
      ()
  | ("bounding_circle_lt", [x; y; r; l]) ->
      add_bounding_circle_lt x y r l
  | ("non_overlap_triangles", [x0; y0; x1; y1; x2; y2; x3; y3; x4; y4; x5; y5]) ->
      add_non_overlap_triangles x0 y0 x1 y1 x2 y2 x3 y3 x4 y4 x5 y5
  | (f, _) ->
      Err.unimplemented_constraint p f

let objective = ref Satisfy

let do_var_decl_annot x a =
  match a with
  | ("output_var", []) ->
      output := OutputVar x :: !output;
  | ("output_array", [Expr.ArrayRange (lb, ub)]) ->
      output := OutputArray (x, lb, ub) :: !output;
  | _ ->
      ()

let rec add_clause clause =
  match clause with
  | PredDecl _ ->
      ()
  | ParDecl (_, x, e) ->
      Hashtbl.add parameters x e
  | VarDecl (t, x, a, e, _) ->
      let v =
        match t with
        | VarType.Array _ ->
            begin
              match e with
              | Some a -> Array a
              | _ -> failwith "add_clause"
            end
        | _ ->
            Unbound in
      Hashtbl.add variables x v;
      List.iter (do_var_decl_annot x) a
  | Cstr (c, a, p) ->
      add_constraint c a p
  | Solve (_, obj) ->
      objective := obj

let search_variables_count = ref 0

let search_variables = Hashtbl.create search_variables_sz

let generated_variables_count = ref 0

let generated_variables = Hashtbl.create generated_variables_sz

let rec generate_expr array chan expr =
  match expr with
  | Expr.Float f ->
      string_of_float f
  | Expr.Var x ->
      begin
	try
	  let result = Hashtbl.find generated_variables x in
	  if result = "" then
	    failwith ("generate_expr " ^ x);
	  result
	with Not_found ->
	  Hashtbl.add generated_variables x "";
	  let s =
	    match Hashtbl.find variables x with
	    | Unbound ->
                let index =
                  try
                    Hashtbl.find search_variables x
                  with Not_found ->
		    let index = !search_variables_count in
                    Hashtbl.add search_variables x index;
		    search_variables_count := index + 1;
                    index in
		Printf.sprintf "%s[%d]" array index
	    | Unary_op (op, x) ->
		let index = !generated_variables_count in
		generated_variables_count := index + 1;
		let v = Printf.sprintf "x%d" index in
		let sx = generate_expr array chan x in
		Printf.fprintf chan "  double %s = %s %s;\n" v op sx;
		v
	    | Bin_op (x, op, y) ->
		let index = !generated_variables_count in
		generated_variables_count := index + 1;
		let v = Printf.sprintf "x%d" index in
		let sx = generate_expr array chan x in
		let sy = generate_expr array chan y in
		Printf.fprintf chan "  double %s = %s %s %s;\n" v sx op sy;
		v
	    | Min (x, y) ->
		let index = !generated_variables_count in
		generated_variables_count := index + 1;
		let v = Printf.sprintf "x%d" index in
		let sx = generate_expr array chan x in
		let sy = generate_expr array chan y in
		Printf.fprintf chan "  double %s = fmin(%s, %s);\n" v sx sy;
		v
	    | Max (x, y) ->
		let index = !generated_variables_count in
		generated_variables_count := index + 1;
		let v = Printf.sprintf "x%d" index in
		let sx = generate_expr array chan x in
		let sy = generate_expr array chan y in
		Printf.fprintf chan "  double %s = fmax(%s, %s);\n" v sx sy;
		v
	    | Sqrt x ->
		let index = !generated_variables_count in
		generated_variables_count := index + 1;
		let v = Printf.sprintf "x%d" index in
		let sx = generate_expr array chan x in
		Printf.fprintf chan "  double %s = sqrt(%s);\n" v sx;
		v
	    | Sin x ->
		let index = !generated_variables_count in
		generated_variables_count := index + 1;
		let v = Printf.sprintf "x%d" index in
		let sx = generate_expr array chan x in
		Printf.fprintf chan "  double %s = sin(%s);\n" v sx;
		v
	    | Cos x ->
		let index = !generated_variables_count in
		generated_variables_count := index + 1;
		let v = Printf.sprintf "x%d" index in
		let sx = generate_expr array chan x in
		Printf.fprintf chan "  double %s = cos(%s);\n" v sx;
		v
            | Lin_eq (vars, constant) ->
		let index = !generated_variables_count in
		generated_variables_count := index + 1;
		let v = Printf.sprintf "x%d" index in
                let vars = List.map (fun (coef, x) -> (coef, generate_expr array chan x)) vars in
		Printf.fprintf chan "  double %s = %f" v constant;
                List.iter begin fun (coef, x) ->
                  Printf.fprintf chan " + %f * %s" coef x
                end vars;
                Printf.fprintf chan ";\n";
		v
            | Array _ ->
                failwith "generate_expr array" in
	  Hashtbl.replace generated_variables x s;
	  s
      end
  | _ ->
      failwith "generate_expr other"

let rec get_expr_list e =
  match e with
  | Expr.Var x ->
      get_expr_list (Hashtbl.find parameters x)
  | Expr.ArrayList v ->
      v
  | _ ->
      failwith "generate_expr_list"

let generate_expr_list chan e =
  List.map (generate_expr "x" chan) (get_expr_list e)

let generate_constraint chan cstr =
  match cstr with
  | Lin_lt (x, y, z) ->
      let sx = generate_expr_list chan x in
      let sy = generate_expr_list chan y in
      let sz = generate_expr "x" chan z in
      Printf.fprintf chan "  cost += constraint_cost(- %s" sz;
      List.iter2 begin fun x y ->
	Printf.fprintf chan " + %s * %s" x y
      end sx sy;
      Printf.fprintf chan ");\n"
  | Lt (x, y) ->
      let sx = generate_expr "x" chan x in
      let sy = generate_expr "x" chan y in
      Printf.fprintf chan "  cost += constraint_cost(%s - %s);\n" sx sy
  | BoudingCircleLt (x, y, r, l) ->
      let sx = generate_expr "x" chan x in
      let sy = generate_expr "x" chan y in
      let sr = generate_expr "x" chan r in
      let sl = generate_expr "x" chan l in
      Printf.fprintf chan "  cost += constraint_cost(sqrt(%s * %s + %s * %s) + %s - %s);\n" sx sx sy sy sr sl
  | NonOverlapTriangles (x0, y0, x1, y1, x2, y2, x3, y3, x4, y4, x5, y5) ->
      let sx0 = generate_expr "x" chan x0 in
      let sy0 = generate_expr "x" chan y0 in
      let sx1 = generate_expr "x" chan x1 in
      let sy1 = generate_expr "x" chan y1 in
      let sx2 = generate_expr "x" chan x2 in
      let sy2 = generate_expr "x" chan y2 in
      let sx3 = generate_expr "x" chan x3 in
      let sy3 = generate_expr "x" chan y3 in
      let sx4 = generate_expr "x" chan x4 in
      let sy4 = generate_expr "x" chan y4 in
      let sx5 = generate_expr "x" chan x5 in
      let sy5 = generate_expr "x" chan y5 in
      Printf.fprintf chan "  cost += constraint_cost(distance_triangles(%s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s));\n" sx0 sy0 sx1 sy1 sx2 sy2 sx3 sy3 sx4 sy4 sx5 sy5

let generate_objective chan obj =
  match obj with
  | Satisfy -> ()
  | Minimize e ->
      let s = generate_expr "x" chan e in
      Printf.fprintf chan "  if (%s > 0) cost += %s;\n" s s
  | Maximize e ->
      let s = generate_expr "x" chan e in
      Printf.fprintf chan "  cost -= %s;\n" s

let generate_fitfun chan =
  Printf.fprintf chan "\
#include <stdio.h>
#include <math.h>
#include \"fitfun.h\"

double fitfun(double const x[]) {
  double cost = 0;
";
  List.iter (generate_constraint chan) !constraints;
  generate_objective chan !objective;
  Printf.fprintf chan "  return cost;
}
"

let generate_output chan o =
  match o with
  | OutputVar x ->
      let sx = generate_expr "x" chan (Expr.Var x) in
      Printf.fprintf chan "  printf(\"%s = %%f;\\n\", %s);\n" x sx
  | OutputArray (x, lb, ub) ->
      let a =
        try Hashtbl.find variables x
        with Not_found -> failwith "array not found" in
      Printf.fprintf chan "  printf(\"%s = array1d(%d..%d, [" x lb ub;
      let l =
        match a with
        | Array l -> l
        | _ -> failwith "generate_output" in
      let first = ref true in
      List.iter begin fun _ ->
        if !first then
          first := false
        else
          Printf.fprintf chan ", ";
        Printf.fprintf chan "%%f"
      end l;
      Printf.fprintf chan "]);\\n\"";
      let l = List.map begin fun v ->
        generate_expr "x" chan v
      end l in
      List.iter begin fun s ->
        Printf.fprintf chan ", %s" s
      end l;
      Printf.fprintf chan ");\n"

let generate_output output_filename =
  write_file output_filename begin fun chan ->
    generate_fitfun chan;
    Printf.fprintf chan "\

const int VARIABLE_COUNT = %d;

" !search_variables_count;
    Hashtbl.clear generated_variables;
  Printf.fprintf chan "\
double print_solution(double const x[]) {
";
    List.iter (generate_output chan) (List.rev !output);
    Printf.fprintf chan "  printf(\"----------\\n\");
  printf(\"==========\\n\");
  return 0;
}
"
  end

let compile filename =
  let clauses = parse filename in
  List.iter add_clause clauses;
  let output_filename = Filename.chop_extension filename ^ ".c" in
  generate_output output_filename

let main () =
  let filename = Sys.argv.(1) in
  compile filename

let () =
  try
    main ()
  with
  | Failure msg
  | Sys_error msg ->
      prerr_endline msg;
      exit 1
  | Err.E (position, msg) ->
      let open Pos in
      if position.start_p.line = position.end_p.line then
        if position.start_p.column = position.end_p.column then
          Format.eprintf "%s:%d.%d: %s@." position.filename position.start_p.line position.start_p.column msg
        else
          Format.eprintf "%s:%d.%d-%d: %s@." position.filename position.start_p.line position.start_p.column position.end_p.column msg
      else
        Format.eprintf "%s:%d.%d-%d.%d: %s@." position.filename position.start_p.line position.start_p.column position.end_p.line position.end_p.column msg;
      exit 1
